from dataclasses import asdict, is_dataclass, dataclass
from enum import Enum
from typing import Awaitable, Callable, Generic
import multiprocessing
from contextlib import asynccontextmanager

from starlette.applications import Starlette
from starlette.responses import JSONResponse
from starlette.requests import Request as StarletteRequest
from starlette.routing import Route

from onbbu.database import DatabaseManager
from onbbu.types import T


class Request(StarletteRequest):
    pass


class ResponseNotFoundError(Generic[T], JSONResponse):

    def render(self, content: T) -> bytes:
        content = {"error": str(content)}  # type: ignore

        return super().render(content, status_code=404)  # type: ignore


class ResponseValueError(Generic[T], JSONResponse):

    def render(self, content: T) -> bytes:
        content = {"error": str(content)}  # type: ignore

        return super().render(content, status_code=500)  # type: ignore


class ResponseValidationError(Generic[T], JSONResponse):

    def render(self, content: T) -> bytes:
        content = {"detail": content.errors()}  # type: ignore

        return super().render(content, status_code=400)  # type: ignore


class Response(Generic[T], JSONResponse):

    def render(self, content: T) -> bytes:  # type: ignore

        if is_dataclass(content):
            content = asdict(content)

        elif isinstance(content, list):
            content = [
                (asdict(item) if is_dataclass(item) else item) for item in content
            ]

        return super().render(content)


class HTTPMethod(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"


EndpointHttpType = Callable[[Request], Awaitable[JSONResponse]]


class RouterHttp:
    __router: list[Route]
    __prefix: str

    def __init__(self, prefix: str = ""):
        self.__router = []
        self.__prefix = prefix.rstrip("/")

    def add_route(self, path: str, endpoint: EndpointHttpType, method: HTTPMethod):

        full_path: str = f"{self.__prefix}{path}"

        self.__router.append(
            Route(path=full_path, endpoint=endpoint, methods=[method.value])
        )

    def get_router(self):
        return self.__router

    def get_routes(self):
        return [route.path for route in self.__router]

    def get_endpoints(self):
        return [route.endpoint for route in self.__router]


class ServerHttp:
    database: DatabaseManager

    def __init__(self, environment: str, database: DatabaseManager, port: int = 8000):
        self.host = "0.0.0.0"
        self.port = port
        self.environment = environment
        self.reload = self.environment == "development"
        self.workers = 1 if self.reload else max(2, multiprocessing.cpu_count() - 1)

        self.server = Starlette(debug=True, routes=[], lifespan=self._lifespan)

        self.config = ConfigInit(http=self)

        self.database = database

    @asynccontextmanager
    async def _lifespan(self, app: Starlette):
        """Gestor de eventos de vida para FastAPI"""
        await self.database.init()
        yield
        await self.database.close()

    def include_router(self, router: RouterHttp):
        """Agrega todas las rutas de un RouterHttp a la aplicación"""
        self.server.router.routes.extend(router.get_router())

@dataclass(frozen=True, slots=True)
class ConfigInit:
    http: ServerHttp


import os
from types import ModuleType
from typing import Optional

from importlib.machinery import ModuleSpec, SourceFileLoader
from importlib.util import spec_from_file_location, module_from_spec

from onbbu.types import T


class ConfigLoader:
    def __init__(self, base_dir: str):
        self.base_dir = base_dir

    def load_python_config(
        self, path: str, attribute: str, default: Optional[T]
    ) -> Optional[T]:
        """Carga un archivo de configuración en formato Python"""

        config_path: str = os.path.join(self.base_dir, *path.split("/"))

        if not os.path.exists(config_path):
            print(
                f"⚠️ Advertencia: No se encontró `{path}`. Se usará el valor por defecto."
            )
            return default

        spec: ModuleSpec | None = spec_from_file_location("config_module", config_path)

        if spec is None or spec.loader is None:
            print(
                f"⚠️ Advertencia: El archivo `{path}` no tiene un cargador de módulo válido."
            )
            return default

        if isinstance(spec.loader, SourceFileLoader):

            config_module: ModuleType = module_from_spec(spec)

            spec.loader.exec_module(config_module)

            return getattr(config_module, attribute, default)

        print(f"⚠️ Advertencia: El cargador de módulo no es válido para `{path}`.")

        return default



from argparse import ArgumentParser, Namespace
import sys
import os

from typing import Any, List
import uvicorn

from onbbu.logger import LogLevel, logger
from onbbu.database import DatabaseManager
from onbbu.manager import BaseCommand, Manager, create_module, COMMANDS, register_command
from onbbu.config_loader import ConfigLoader
from onbbu.paginate import PaginateDTO, createPaginateResponse, Paginate
from onbbu.http import (
    HTTPMethod,
    EndpointHttpType,
    Response,
    ResponseNotFoundError,
    ResponseValueError,
    ResponseValidationError,
    Request,
    JSONResponse,
    ServerHttp,
)

BASE_DIR: str = os.getcwd()

sys.path.append(BASE_DIR)

environment: str = os.getenv("ENVIRONMENT", "development")

installed_apps: List[str] = ConfigLoader(BASE_DIR).load_python_config(
    path="internal/settings.py",
    attribute="INSTALLED_APPS",
    default=[],
) # type: ignore

server: ServerHttp = ConfigLoader(BASE_DIR).load_python_config(
    path="internal/main.py",
    attribute="server",
    default=None,
) # type: ignore


database = DatabaseManager(
    database_url=os.getenv("DATABASE_URL", "sqlite://db.sqlite3"),
    INSTALLED_APPS=installed_apps,
)

def create_app(port: int = 8000) -> ServerHttp:
    """Crea y retorna una instancia del servidor."""
    return ServerHttp(port=port, database=database, environment=environment)


@register_command
class CommandMigrate(BaseCommand):
    """Command to run the server."""

    name: str = "migrate"
    help: str = "Command to run the server."

    # def add_arguments(self, parser):
    #    parser.add_argument(
    #        "--host", type=str, default="0.0.0.0", help="Host for the server"
    #    )
    #    parser.add_argument(
    #        "--port", type=int, default=8000, help="Port for the server"
    #    )

    async def handle(self, *args: Namespace, **kwargs: dict[str, Any]) -> None:

        await database.init()

        await database.migrate()

        await database.close()


@register_command
class CommandCreateModule(BaseCommand):
    """Command to run the server."""

    name: str = "create_module"
    help: str = "Comando para crear un nuevo modulo."

    async def add_arguments(self, parser: ArgumentParser):
        parser.add_argument(
            "--name", type=str, default="demo", help="Nombre del moodulo"
        )

    async def handle(self, *args: Namespace, **kwargs: dict[str, Any]) -> None:
        await create_module(path=BASE_DIR, name=args.name) # type: ignore


@register_command
class CommandRunServer(BaseCommand):
    """Command to run the server."""

    name: str = "run"
    help: str = "Command to run the server."

    # def add_arguments(self, parser):
    #    parser.add_argument(
    #        "--host", type=str, default="0.0.0.0", help="Host for the server"
    #    )
    #    parser.add_argument(
    #        "--port", type=int, default=8000, help="Port for the server"
    #    )

    async def handle(self, *args: Namespace, **kwargs: dict[str, Any]) -> None:

        if hasattr(server, "server"):
            logger.log(
                level=LogLevel.INFO,
                message=f"🚀 Iniciando servidor en {server.host}:{server.port} ...",
                extra_data={},
            )

            for route in server.server.routes:
                logger.log(
                    level=LogLevel.INFO,
                    message=f"🔗 {route.path} -> {route.name} ({route.methods})", # type: ignore
                    extra_data={},
                )

            uvicorn.run(
                "internal.main:server.server",
                host=server.host,
                port=server.port,
                reload=server.reload,
                workers=server.workers,
            )

        else:
            logger.log(
                level=LogLevel.ERROR,
                message=f"❌ `internal/main.py` no contiene una instancia `server`.",
                extra_data={},
            )


async def cli():

    manager: Manager = Manager(INSTALLED_APPS=installed_apps, COMMANDS=COMMANDS)

    commands: list[BaseCommand] = [
        CommandRunServer(),
        CommandMigrate(),
        CommandCreateModule(),
    ]

    await manager.load_commands()

    await manager.internal_command(commands)

    await manager.execute()


#def main():
#
#    parser: ArgumentParser = ArgumentParser(description="CLI para manejar Onbbu")
#
#    subparsers = parser.add_subparsers(dest="command")
#
#    subparsers.add_parser("run", help="Iniciar el servidor")
#
#    subparsers.add_parser("makemigrations", help="Generar una nueva migración")
#
#    subparsers.add_parser("migrate", help="Aplicar migraciones")
#
#    subparsers.add_parser("routes", help="Listar rutas de")
#
#    args: Namespace = parser.parse_args()
#
#    parser.print_help()
#
#    # if args.command == "run":
#    #    run_server(server=server)
#    # elif args.command == "routes":
#    #    list_routes()
#    # else:
#    #    parser.print_help()


__all__ = [
    "COMMANDS",
    "BaseCommand",
    "register_command",
    "DatabaseManager",
    "HTTPMethod",
    "EndpointHttpType",
    "Paginate",
    "PaginateDTO",
    "createPaginateResponse",
    "Request",
    "JSONResponse",
    "Response",
    "ResponseNotFoundError",
    "ResponseValidationError",
    "ResponseValueError",
]



from collections import defaultdict
from enum import Enum
from typing import Callable, List

class Event(Enum):
    DEFAULT_DISCOUNT = "add_default_discounts"

class EventEmitter:
    def __init__(self) -> None:
        self.events: defaultdict[str, List[Callable[..., None]]] = defaultdict(list)

    def on(self, event: Event, callback: Callable[..., None]) -> None:
        self.events[event.value].append(callback)

    def emit(self, event: Event, *args: object, **kwargs: object) -> None:
        for callback in self.events[event.value]:
            callback(*args, **kwargs)

emitter: EventEmitter = EventEmitter()



from argparse import ArgumentParser, HelpFormatter, Namespace
import importlib
import os
import pkgutil
from typing import Any, Dict, List, Optional


async def create_module(path: str, name: str):

    folders: list[str] = [
        "domain",
        "domain/entities",
        "domain/services",
        "application/dto",
        "application/commands",
        "application/use_cases",
        "infrastructure/adapters",
        "infrastructure/cache",
        "infrastructure/logger",
        "infrastructure/messaging",
        "infrastructure/persistence/models",
        "infrastructure/persistence/repositories",
        "infrastructure/services",
        "infrastructure/storage",
        "infrastructure/transformers",
    ]

    for folder in folders:
        dir_path = os.path.join(path, "pkg", name, folder)
        os.makedirs(dir_path, exist_ok=True)


class BaseCommand:
    """Base class for all commands."""

    name: str
    help: str = "Base command description"
    parser: Optional[ArgumentParser] = None

    def __init__(self) -> None:
        self.parser: Optional[ArgumentParser] = None

    async def add_arguments(self, parser: ArgumentParser) -> None:
        """Override this method to add custom arguments."""
        pass

    async def handle(self, *args: Namespace, **kwargs: dict[str, Any]) -> None:
        """Override this method to implement command logic."""
        raise NotImplementedError("Subclasses must implement handle()")


COMMANDS: Dict[str, BaseCommand] = {}


def register_command(cls: type[BaseCommand]) -> type[BaseCommand]:
    """Decorator to register commands automatically."""
    COMMANDS[cls.name.lower()] = cls()
    return cls


class Manager:
    """Command-line manager for executing various tasks."""

    def __init__(self, INSTALLED_APPS: List[str], COMMANDS: Dict[str, BaseCommand]):
        self.parser = ArgumentParser(
            description="Management script",
            formatter_class=lambda prog: HelpFormatter(prog, max_help_position=30),
        )

        self.subparsers = self.parser.add_subparsers(dest="command", metavar="command")

        self.INSTALLED_APPS = INSTALLED_APPS

        self.COMMANDS = COMMANDS

    async def internal_command(self, commands: list[BaseCommand]):
        """Execute a system command."""

        for command in commands:
            self.COMMANDS[command.name] = command

    async def load_commands(self):
        """Dynamically load commands from installed apps."""
        for app in self.INSTALLED_APPS:
            commands_path = f"pkg.{app}.application.commands"
            try:
                module = importlib.import_module(commands_path)

                if hasattr(module, "__path__"):
                    for _, module_name, _ in pkgutil.iter_modules(module.__path__):

                        full_module_name = f"{commands_path}.{module_name}"

                        command_module = importlib.import_module(full_module_name)

                        if hasattr(command_module, "Command"):
                            command_instance: BaseCommand = command_module.Command()
                            self.COMMANDS[command_instance.name] = command_instance

            except ModuleNotFoundError as e:
                print(f"Warning: Could not import {commands_path}: {e}")

        for name, command in self.COMMANDS.items():

            command_parser = self.subparsers.add_parser(name, help=command.help)

            await command.add_arguments(command_parser)

            command.parser = command_parser

    async def execute(self):
        """Parse and execute commands dynamically."""

        args: Namespace = self.parser.parse_args()

        command = self.COMMANDS.get(args.command)

        if command:
            await command.handle(args)
        else:
            self.parser.print_help()
